@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2009,2011,2012,2013,2014 Red Hat, Inc.
+ * Copyright (C) 2009,2011,2012,2013,2014,2015 Red Hat, Inc.
  * 
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -53,7 +53,8 @@
 #include "store-int.h"
 #include "util-n.h"
 
-/* Structures and templates for parsing principal name otherName values. */
+/* Structures and templates for creating and parsing principal name otherName
+ * values. */
 struct realm {
 	SECItem name;
 };
@@ -315,26 +316,26 @@ cm_certext_read_ku(struct cm_store_entry *entry, PLArenaPool *arena,
 	}
 }
 
-/* Build a keyUsage extension value from a string, with each bit being
+/* Build a BitString extension value from a string, with each bit being
  * represented by either a "1" or a "0", most significant bit first. */
 static SECItem *
-cm_certext_build_ku(struct cm_store_entry *entry, PLArenaPool *arena,
-		    const char *ku_value)
+cm_certext_build_bitstring(struct cm_store_entry *entry, PLArenaPool *arena,
+			   const char *bitstring)
 {
 	SECItem *ret, encoded, *bits;
 	unsigned int i, used, val, len;
 
-	if ((ku_value == NULL) || (strlen(ku_value) == 0)) {
+	if ((bitstring == NULL) || (strlen(bitstring) == 0)) {
 		/* Nothing to encode, so don't include this extension. */
 		return NULL;
 	}
-	len = strlen(ku_value) + 1;
+	len = strlen(bitstring) + 1;
 	bits = SECITEM_AllocItem(arena, NULL, len);
 	memset(bits->data, '\0', len);
 	for (i = 0, used = 0;
-	     (ku_value != NULL) && (ku_value[i] != '\0');
+	     (bitstring != NULL) && (bitstring[i] != '\0');
 	     i++) {
-		val = ((ku_value[i] == '1') ? 0x80 : 0x00) >> (i % 8);
+		val = ((bitstring[i] == '1') ? 0x80 : 0x00) >> (i % 8);
 		bits->data[i / 8] |= val;
 		if (val != 0) {
 			used = i + 1;
@@ -353,6 +354,15 @@ cm_certext_build_ku(struct cm_store_entry *entry, PLArenaPool *arena,
 	return ret;
 }
 
+/* Build a keyUsage extension value from a string, with each bit being
+ * represented by either a "1" or a "0", most significant bit first. */
+static SECItem *
+cm_certext_build_ku(struct cm_store_entry *entry, PLArenaPool *arena,
+		    const char *ku_value)
+{
+	return cm_certext_build_bitstring(entry, arena, ku_value);
+}
+
 /* Convert an OID to a printable string.  For now, we're limited to components
  * that will fit into a "long". */
 static char *
@@ -573,6 +583,8 @@ cm_certext_build_eku(struct cm_store_entry *entry, PLArenaPool *arena,
 	return ret;
 }
 
+/* Pull the nth component out of a principal name structure.  Treat numbers
+ * less than zero as a request for the realm name. */
 static unsigned char *
 cm_certext_princ_data(krb5_context ctx, krb5_principal princ, int i)
 {
@@ -591,6 +603,8 @@ cm_certext_princ_data(krb5_context ctx, krb5_principal princ, int i)
 	}
 }
 
+/* Return the length of the data that cm_certext_princ_data() will return for a
+ * given index. */
 static int
 cm_certext_princ_len(krb5_context ctx, krb5_principal princ, int i)
 {
@@ -609,6 +623,7 @@ cm_certext_princ_len(krb5_context ctx, krb5_principal princ, int i)
 	}
 }
 
+/* Return a the name-type from a principal name structure. */
 static int
 cm_certext_princ_get_type(krb5_context ctx, krb5_principal princ)
 {
@@ -619,6 +634,7 @@ cm_certext_princ_get_type(krb5_context ctx, krb5_principal princ)
 #endif
 }
 
+/* Set the name-type in a principal name structure. */
 static void
 cm_certext_princ_set_type(krb5_context ctx, krb5_principal princ, int nt)
 {
@@ -629,6 +645,7 @@ cm_certext_princ_set_type(krb5_context ctx, krb5_principal princ, int nt)
 #endif
 }
 
+/* Free an unparsed principal name. */
 static void
 cm_certext_free_unparsed_name(krb5_context ctx, char *name)
 {
@@ -639,6 +656,7 @@ cm_certext_free_unparsed_name(krb5_context ctx, char *name)
 #endif
 }
 
+/* Check how many components are in a principal name. */
 static int
 cm_certext_princ_get_length(krb5_context ctx, krb5_principal princ)
 {
@@ -649,6 +667,7 @@ cm_certext_princ_get_length(krb5_context ctx, krb5_principal princ)
 #endif
 }
 
+/* Set how many components are in a principal name. */
 static void
 cm_certext_princ_set_length(krb5_context ctx, krb5_principal princ, int length)
 {
@@ -659,6 +678,8 @@ cm_certext_princ_set_length(krb5_context ctx, krb5_principal princ, int length)
 #endif
 }
 
+/* Set a realm name in a principal name to point to a copy of the passed-in
+ * name owned by "parent". */
 static void
 cm_certext_princ_set_realm(krb5_context ctx, void *parent, krb5_principal princ,
 			   int length, char *name)
@@ -676,6 +697,8 @@ cm_certext_princ_set_realm(krb5_context ctx, void *parent, krb5_principal princ,
 #endif
 }
 
+/* Append a component to a principal name, using storage owned by "parent" to
+ * hold a copy of the passed-in component value. */
 static void
 cm_certext_princ_append_comp(krb5_context ctx, void *parent,
 			     krb5_principal princ, char *name, int length)
@@ -717,16 +740,18 @@ static char *
 cm_certext_parse_principal(void *parent, struct kerberos_principal_name *p)
 {
 	SECItem **comps;
+	krb5_error_code err;
 	krb5_context ctx;
 	krb5_principal_data princ;
 	char *unparsed, *ret;
 	int i, j;
 	unsigned long name_type;
 	void *tctx;
+
 	ret = NULL;
 	ctx = NULL;
 	tctx = talloc_new(parent);
-	if (krb5_init_context(&ctx) == 0) {
+	if ((err = krb5_init_context(&ctx)) == 0) {
 		memset(&princ, 0, sizeof(princ));
 		/* Copy the realm over. */
 		cm_certext_princ_set_realm(ctx, tctx, &princ,
@@ -760,6 +785,9 @@ cm_certext_parse_principal(void *parent, struct kerberos_principal_name *p)
 		}
 		talloc_free(tctx);
 		krb5_free_context(ctx);
+	} else {
+		cm_log(1, "Error %ld initializing Kerberos.\n",
+		       (long) err);
 	}
 	return ret;
 }
@@ -1003,6 +1031,7 @@ cm_certext_build_principal(struct cm_store_entry *entry, PLArenaPool *arena,
 {
 	SECItem *comp, **comps, encoded;
 	struct kerberos_principal_name p;
+	krb5_error_code err;
 	krb5_context ctx;
 	krb5_principal princ;
 	int i;
@@ -1011,7 +1040,9 @@ cm_certext_build_principal(struct cm_store_entry *entry, PLArenaPool *arena,
 		return NULL;
 	}
 	ctx = NULL;
-	if (krb5_init_context(&ctx) != 0) {
+	if ((err = krb5_init_context(&ctx)) != 0) {
+		cm_log(1, "Error %ld initializing Kerberos.\n",
+		       (long) err);
 		return NULL;
 	}
 	princ = NULL;
@@ -1445,7 +1476,7 @@ cm_certext_build_aia(struct cm_store_entry *entry, PLArenaPool *arena,
 	return item;
 }
 
-/* Build a CRL distribution points extension value. */
+/* Build a CRL distribution points or freshest CRL extension value. */
 static SECItem *
 cm_certext_build_crldp(struct cm_store_entry *entry, PLArenaPool *arena,
 		       char **crldp)
@@ -1454,13 +1485,8 @@ cm_certext_build_crldp(struct cm_store_entry *entry, PLArenaPool *arena,
 	CRLDistributionPoint *value, **values;
 	CERTGeneralName *location;
 	SECItem encoded, *item;
-	SECOidData *oid;
 	unsigned int i, j, n;
 
-	oid = SECOID_FindOIDByTag(SEC_OID_PKIX_OCSP);
-	if (oid == NULL) {
-		return NULL;
-	}
 	for (n = 0; (crldp != NULL) && (crldp[n] != NULL); n++) {
 		continue;
 	}
@@ -1546,6 +1572,86 @@ cm_certext_build_ocsp_no_check(struct cm_store_entry *entry,
 	return item;
 }
 
+/* Build a Microsoft certtype extension value. */
+static SECItem *
+cm_certext_build_profile(struct cm_store_entry *entry,
+			 PLArenaPool *arena,
+			 char *profile)
+{
+	SECItem value, encoded, *item;
+	unsigned int len = 0;
+
+	if (strlen(profile) == 0) {
+		return NULL;
+	}
+	memset(&value, 0, sizeof(value));
+	memset(&encoded, 0, sizeof(encoded));
+	if (cm_store_utf8_to_bmp_string(profile, &value.data, &len) != -1) {
+		value.len = len;
+		if (SEC_ASN1EncodeItem(arena, &encoded, &value,
+				       SEC_BMPStringTemplate) == &encoded) {
+			item = SECITEM_ArenaDupItem(arena, &encoded);
+		} else {
+			item = NULL;
+		}
+		free(value.data);
+	} else {
+		item = NULL;
+	}
+	return item;
+}
+
+/* Build a Netscape certtype extension value. */
+static SECItem *
+cm_certext_build_ns_certtype(struct cm_store_entry *entry,
+			     PLArenaPool *arena,
+			     char *certtype)
+{
+	char bitstring[] = "00000000";
+	char *p, *q;
+	int len = 0;
+
+	if (strlen(certtype) == 0) {
+		return NULL;
+	}
+	p = certtype;
+	while (*p != '\0') {
+		q = p + strcspn(p, ",");
+		if (strncasecmp(p, "client", q - p) == 0) {
+			bitstring[0] = '1';
+		} else
+		if (strncasecmp(p, "server", q - p) == 0) {
+			bitstring[1] = '1';
+		} else
+		if (strncasecmp(p, "email", q - p) == 0) {
+			bitstring[2] = '1';
+		} else
+		if (strncasecmp(p, "objsign", q - p) == 0) {
+			bitstring[3] = '1';
+		} else
+		if (strncasecmp(p, "reserved", q - p) == 0) {
+			bitstring[4] = '1';
+		} else
+		if (strncasecmp(p, "sslca", q - p) == 0) {
+			bitstring[5] = '1';
+		} else
+		if (strncasecmp(p, "emailca", q - p) == 0) {
+			bitstring[6] = '1';
+		} else
+		if (strncasecmp(p, "objca", q - p) == 0) {
+			bitstring[7] = '1';
+		}
+		p = q + strspn(q, ",");
+	}
+	if (strchr(bitstring, '1') != NULL) {
+		len = strrchr(bitstring, '1') - bitstring;
+		p[len + 1] = '\0';
+		return cm_certext_build_bitstring(entry, arena, bitstring);
+	} else {
+		return NULL;
+	}
+}
+
 /* Build a requestedExtensions attribute. */
 void
 cm_certext_build_csr_extensions(struct cm_store_entry *entry,
@@ -1553,7 +1659,7 @@ cm_certext_build_csr_extensions(struct cm_store_entry *entry,
 				unsigned char **extensions, size_t *length)
 {
 	PLArenaPool *arena;
-	CERTCertExtension ext[11], *exts[12], **exts_ptr;
+	CERTCertExtension ext[13], *exts[14], **exts_ptr;
 	SECOidData *oid;
 	SECItem *item, encoded;
 	SECItem der_false = {
@@ -1730,6 +1836,30 @@ cm_certext_build_csr_extensions(struct cm_store_entry *entry,
 			i++;
 		}
 	}
+	if (entry->cm_template_profile != NULL) {
+		oid = (SECOidData *) &oid_microsoft_certtype;
+		item = cm_certext_build_profile(entry, arena,
+						entry->cm_template_profile);
+		if ((item != NULL) && (oid != NULL)) {
+			ext[i].id = oid->oid;
+			ext[i].critical = der_false;
+			ext[i].value = *item;
+			exts[i] = &ext[i];
+			i++;
+		}
+	}
+	if (entry->cm_template_ns_certtype != NULL) {
+		oid = SECOID_FindOIDByTag(SEC_OID_NS_CERT_EXT_CERT_TYPE);
+		item = cm_certext_build_ns_certtype(entry, arena,
+						    entry->cm_template_ns_certtype);
+		if ((item != NULL) && (oid != NULL)) {
+			ext[i].id = oid->oid;
+			ext[i].critical = der_false;
+			ext[i].value = *item;
+			exts[i] = &ext[i];
+			i++;
+		}
+	}
 	exts[i++] = NULL;
 	exts_ptr = exts;
 	/* Encode the sequence. */
@@ -1758,6 +1888,7 @@ cm_certext_build_csr_extensions(struct cm_store_entry *entry,
 	PORT_FreeArena(arena, PR_TRUE);
 }
 
+/* Read a basicConstraints extension. */
 static void
 cm_certext_read_basic(struct cm_store_entry *entry, PLArenaPool *arena,
 		      CERTCertExtension *ext)
@@ -1776,6 +1907,7 @@ cm_certext_read_basic(struct cm_store_entry *entry, PLArenaPool *arena,
 	}
 }
 
+/* Read a Netscape comment extension. */
 static void
 cm_certext_read_nsc(struct cm_store_entry *entry, PLArenaPool *arena,
 		    CERTCertExtension *ext)
@@ -1798,6 +1930,8 @@ cm_certext_read_nsc(struct cm_store_entry *entry, PLArenaPool *arena,
 	}
 }
 
+/* Read an authorityInformationAccess extension, and keep track of any OCSP
+ * responders that we find in it. */
 static void
 cm_certext_read_aia(struct cm_store_entry *entry, PLArenaPool *arena,
 		    CERTCertExtension *ext)
@@ -1842,6 +1976,8 @@ cm_certext_read_aia(struct cm_store_entry *entry, PLArenaPool *arena,
 	}
 }
 
+/* Read a CRL distribution points or freshest CRL extension, and return any
+ * locations that we find in it. */
 static void
 cm_certext_read_crlext(struct cm_store_entry *entry, PLArenaPool *arena,
 		       CERTCertExtension *ext, char ***dest)
@@ -1889,6 +2025,7 @@ cm_certext_read_crlext(struct cm_store_entry *entry, PLArenaPool *arena,
 	*dest = list;
 }
 
+/* Read the list of CRL distribution points. */
 static void
 cm_certext_read_crldp(struct cm_store_entry *entry, PLArenaPool *arena,
 		      CERTCertExtension *ext)
@@ -1897,6 +2034,7 @@ cm_certext_read_crldp(struct cm_store_entry *entry, PLArenaPool *arena,
 			       &entry->cm_cert_crl_distribution_point);
 }
 
+/* Read the list of locations where we can find the freshest CRL. */
 static void
 cm_certext_read_freshest_crl(struct cm_store_entry *entry, PLArenaPool *arena,
 			     CERTCertExtension *ext)
@@ -1904,6 +2042,7 @@ cm_certext_read_freshest_crl(struct cm_store_entry *entry, PLArenaPool *arena,
 	cm_certext_read_crlext(entry, arena, ext, &entry->cm_cert_freshest_crl);
 }
 
+/* Parse the data from a Microsoft certificate type extension. */
 static void
 cm_certext_read_profile(struct cm_store_entry *entry, PLArenaPool *arena,
 			CERTCertExtension *ext)
@@ -1928,6 +2067,77 @@ cm_certext_read_profile(struct cm_store_entry *entry, PLArenaPool *arena,
 	}
 }
 
+/* Parse the data from a Netscape certificate type extension. */
+static void
+cm_certext_read_ns_certtype(struct cm_store_entry *entry, PLArenaPool *arena,
+			    CERTCertExtension *ext)
+{
+	SECItem item;
+	unsigned int i, bit;
+	char *tmp = NULL, *t = NULL;
+
+	if (SEC_ASN1DecodeItem(arena, &item, SEC_BitStringTemplate,
+			       &ext->value) == SECSuccess) {
+		/* A bitString decodes with length == number of bits, not
+		 * bytes, which is what we want anyway. */
+		tmp = talloc_zero_size(entry, item.len + 1);
+		for (i = 0; i < item.len; i++) {
+			bit = (item.data[i / 8] & (0x80 >> (i % 8))) ? 1 : 0;
+			sprintf(tmp + i, "%.*u", 1, bit);
+		}
+	}
+	talloc_free(entry->cm_cert_ns_certtype);
+	entry->cm_cert_ns_certtype = NULL;
+	if (tmp == NULL) {
+		return;
+	}
+	t = talloc_strdup(entry, "");
+	if ((tmp != NULL) && (strlen(tmp) > 0)) {
+		if (tmp[0] == '1') {
+			t = talloc_strdup_append(t, ",client");
+		}
+	}
+	if ((tmp != NULL) && (strlen(tmp) > 1)) {
+		if (tmp[1] == '1') {
+			t = talloc_strdup_append(t, ",server");
+		}
+	}
+	if ((tmp != NULL) && (strlen(tmp) > 2)) {
+		if (tmp[2] == '1') {
+			t = talloc_strdup_append(t, ",email");
+		}
+	}
+	if ((tmp != NULL) && (strlen(tmp) > 3)) {
+		if (tmp[3] == '1') {
+			t = talloc_strdup_append(t, ",objsign");
+		}
+	}
+	if ((tmp != NULL) && (strlen(tmp) > 4)) {
+		if (tmp[4] == '1') {
+			t = talloc_strdup_append(t, ",reserved");
+		}
+	}
+	if ((tmp != NULL) && (strlen(tmp) > 5)) {
+		if (tmp[5] == '1') {
+			t = talloc_strdup_append(t, ",sslCA");
+		}
+	}
+	if ((tmp != NULL) && (strlen(tmp) > 6)) {
+		if (tmp[6] == '1') {
+			t = talloc_strdup_append(t, ",emailCA");
+		}
+	}
+	if ((tmp != NULL) && (strlen(tmp) > 7)) {
+		if (tmp[7] == '1') {
+			t = talloc_strdup_append(t, ",objCA");
+		}
+	}
+	if (strlen(t) > 0) {
+		entry->cm_cert_ns_certtype = talloc_strdup(entry, t + 1);
+	}
+	talloc_free(t);
+}
+
 /* Read the extensions from a certificate. */
 void
 cm_certext_read_extensions(struct cm_store_entry *entry, PLArenaPool *arena,
@@ -1937,7 +2147,7 @@ cm_certext_read_extensions(struct cm_store_entry *entry, PLArenaPool *arena,
 	PLArenaPool *local_arena;
 	SECOidData *ku_oid, *eku_oid, *san_oid, *freshest_crl_oid;
 	SECOidData *basic_oid, *nsc_oid, *aia_oid, *crldp_oid, *profile_oid;
-	SECOidData *no_ocsp_check_oid;
+	SECOidData *no_ocsp_check_oid, *ns_certtype_oid;
 
 	if (extensions == NULL) {
 		return;
@@ -2006,6 +2216,12 @@ cm_certext_read_extensions(struct cm_store_entry *entry, PLArenaPool *arena,
 		return;
 	}
 	profile_oid = (SECOidData *) &oid_microsoft_certtype;
+	ns_certtype_oid = SECOID_FindOIDByTag(SEC_OID_NS_CERT_EXT_CERT_TYPE);
+	if (ns_certtype_oid == NULL) {
+		cm_log(1, "Internal library error: unable to look up OID for "
+		       "nsCertType extension.\n");
+		return;
+	}
 	entry->cm_cert_no_ocsp_check = FALSE;
 	for (i = 0; extensions[i] != NULL; i++) {
 		if (SECITEM_ItemsAreEqual(&ku_oid->oid, &extensions[i]->id)) {
@@ -2043,6 +2259,10 @@ cm_certext_read_extensions(struct cm_store_entry *entry, PLArenaPool *arena,
 					  &extensions[i]->id)) {
 			entry->cm_cert_no_ocsp_check = TRUE;
 		}
+		if (SECITEM_ItemsAreEqual(&ns_certtype_oid->oid,
+					  &extensions[i]->id)) {
+			cm_certext_read_ns_certtype(entry, arena, extensions[i]);
+		}
 	}
 	if (arena == local_arena) {
 		PORT_FreeArena(local_arena, PR_TRUE);
