@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2009,2010,2011,2012,2013,2014 Red Hat, Inc.
+ * Copyright (C) 2009,2010,2011,2012,2013,2014,2015 Red Hat, Inc.
  * 
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -25,11 +25,13 @@
 #include <limits.h>
 #include <stdio.h>
 #include <stdlib.h>
+#include <time.h>
 #include <unistd.h>
 
 #include <nss.h>
 #include <pk11pqg.h>
 #include <pk11pub.h>
+#include <cert.h>
 #include <keyhi.h>
 #include <keythi.h>
 #include <prerror.h>
@@ -50,6 +52,7 @@
 
 struct cm_keygen_state {
 	struct cm_keygen_state_pvt pvt;
+	struct cm_store_entry *entry;
 	struct cm_subproc_state *subproc;
 };
 struct cm_keygen_n_settings {
@@ -76,6 +79,39 @@ pqg_size(int key_size)
 }
 #endif
 
+static char *
+make_nickname(const char *prefix, char **marker)
+{
+	unsigned char suffix[6];
+	char *ret;
+	size_t l;
+
+	if (PK11_GenerateRandom(suffix, sizeof(suffix)) != SECSuccess) {
+		/* Try again sometime later. */
+		cm_log(1, "Error generating suffix: %s.\n",
+		       PR_ErrorToName(PORT_GetError()));
+		_exit(CM_SUB_STATUS_INTERNAL_ERROR);
+	}
+	*marker = cm_store_base64_from_bin(NULL, suffix, sizeof(suffix));
+	if (*marker == NULL) {
+		/* Try again sometime later. */
+		cm_log(1, "Error generating suffix.\n");
+		_exit(CM_SUB_STATUS_INTERNAL_ERROR);
+	}
+	while ((l = strcspn(*marker, "+/")) != strlen(*marker)) {
+		switch ((*marker)[l]) {
+		case '+':
+			(*marker)[l] = '=';
+			break;
+		case '/':
+			(*marker)[l] = '_';
+			break;
+		}
+	}
+	ret = util_build_next_nickname(prefix, *marker);
+	return ret;
+}
+
 static int
 cm_keygen_n_main(int fd, struct cm_store_ca *ca, struct cm_store_entry *entry,
 		 void *userdata)
@@ -103,11 +139,15 @@ cm_keygen_n_main(int fd, struct cm_store_ca *ca, struct cm_store_entry *entry,
 	SECOidData *ecurve;
 	SECItem ec_params;
 #endif
-	SECKEYPrivateKey *privkey, *delkey;
+	SECKEYPrivateKey *privkey, *delkey, *ckey;
 	SECKEYPrivateKeyList *privkeys;
 	SECKEYPrivateKeyListNode *node;
 	SECKEYPublicKey *pubkey;
+	CERTCertList *certs;
+	CERTCertListNode *cnode;
+	CERTCertificate *cert;
 	const char *es, *token, *keyname, *reason;
+	char *nickname, *marker = "", *markertmp;
 	char *pin, *pubhex, *pubihex;
 	struct cm_keygen_n_settings *settings;
 	struct cm_pin_cb_data cb_data;
@@ -162,7 +202,7 @@ cm_keygen_n_main(int fd, struct cm_store_ca *ca, struct cm_store_entry *entry,
 		}
 		if (es != NULL) {
 			fprintf(status, "Error initializing database "
-			        "'%s': %s.\n",
+				"'%s': %s.\n",
 				entry->cm_key_storage_location, es);
 			cm_log(1, "Error initializing database '%s': %s.\n",
 			       entry->cm_key_storage_location, es);
@@ -191,11 +231,10 @@ cm_keygen_n_main(int fd, struct cm_store_ca *ca, struct cm_store_entry *entry,
 	if (cm_key_algorithm == cm_key_unspecified) {
 		cm_key_algorithm = CM_DEFAULT_PUBKEY_TYPE;
 	}
-	cm_key_size = entry->cm_key_type.cm_key_gen_size;
-	if (cm_key_size <= 0) {
-		cm_key_size = CM_DEFAULT_PUBKEY_SIZE;
-	}
 	cm_requested_key_size = entry->cm_key_type.cm_key_gen_size;
+	if (cm_requested_key_size <= 0) {
+		cm_requested_key_size = CM_DEFAULT_PUBKEY_SIZE;
+	}
 	/* Convert our key type to a mechanism. */
 	switch (cm_key_algorithm) {
 	case cm_key_rsa:
@@ -284,9 +323,20 @@ next_slot:
 			}
 			PK11_InitPin(slot, NULL, pin ? pin : "");
 			ec = PORT_GetError();
+			if (ec != 0) {
+				es = PR_ErrorToName(ec);
+			} else {
+				es = NULL;
+			}
 			if (PK11_NeedUserInit(slot)) {
-				cm_log(1, "Key generation slot still "
-				       "needs user PIN to be set.\n");
+				if (es != NULL) {
+					cm_log(1, "Key generation slot still "
+					       "needs user PIN to be set: "
+					       "%s.\n", es);
+				} else {
+					cm_log(1, "Key generation slot still "
+					       "needs user PIN to be set.\n");
+				}
 				PK11_FreeSlotList(slotlist);
 				error = NSS_ShutdownContext(ctx);
 				if (error != SECSuccess) {
@@ -403,6 +453,7 @@ retry_gen:
 				cm_log(1, "Trying again.\n");
 				pqg_params = NULL;
 				pqg_verify = NULL;
+				goto retry_gen;
 			}
 			if (PK11_PQG_VerifyParams(pqg_params, pqg_verify,
 						  &pqg_ok) != SECSuccess) {
@@ -428,6 +479,13 @@ retry_gen:
 					_exit(CM_SUB_STATUS_INTERNAL_ERROR);
 				}
 			}
+			generated_size = pqg_params->prime.len * 8;
+			if ((generated_size < cm_key_size) &&
+			    (generated_size > (cm_key_size * 9 / 10))) {
+				cm_log(1, "Params are a bit small (%d vs %d).  Retrying.\n",
+				       pqg_params->prime.len * 8, cm_key_size);
+				goto retry_gen;
+			}
 			if (pqg_ok == SECFailure) {
 				cm_log(1, "Params are bad.  Retrying.\n");
 			}
@@ -483,15 +541,6 @@ retry_gen:
 	pubkey = NULL;
 	privkey = PK11_GenerateKeyPair(slot, mech, params, &pubkey,
 				       PR_TRUE, PR_TRUE, NULL);
-	/* If we're just a bit(s?) short (as opposed to cut off at an arbitrary
-	 * limit that's less than 90% of what we asked for), try again. */
-	generated_size = SECKEY_PublicKeyStrengthInBits(pubkey);
-	if ((generated_size < cm_key_size) &&
-	    (generated_size > (cm_key_size * 9 / 10))) {
-		cm_log(1, "Ended up with %d instead of %d.  Retrying.\n",
-		       SECKEY_PublicKeyStrengthInBits(pubkey), cm_key_size);
-		goto retry_gen;
-	}
 	/* Retry with the optimum key size. */
 	if (privkey == NULL) {
 		cm_key_size = PK11_GetBestKeyLength(slot, pmech);
@@ -521,37 +570,98 @@ retry_gen:
 			break;
 		}
 	}
-	/* Try to remove any keys with conflicting names. */
-	privkeys = PK11_ListPrivKeysInSlot(slot, entry->cm_key_nickname, NULL);
+	/* If we're just a bit(s?) short (as opposed to cut off at an arbitrary
+	 * limit that's less than 90% of what we asked for), try again. */
+	generated_size = SECKEY_PublicKeyStrengthInBits(pubkey);
+	if ((generated_size < cm_key_size) &&
+	    (generated_size > (cm_key_size * 9 / 10))) {
+		cm_log(1, "Ended up with %d instead of %d.  Retrying.\n",
+		       SECKEY_PublicKeyStrengthInBits(pubkey), cm_key_size);
+		goto retry_gen;
+	}
+	/* Check for keys with the desired name, selecting a new name if
+	 * there's already one with the desired name. */
+	nickname = strdup(entry->cm_key_nickname);
+	if (nickname == NULL) {
+		cm_log(1, "Out of memory.\n");
+		_exit(CM_SUB_STATUS_INTERNAL_ERROR);
+	}
+	privkeys = PK11_ListPrivKeysInSlot(slot, nickname, NULL);
 	while ((privkeys != NULL) && !PRIVKEY_LIST_EMPTY(privkeys)) {
-		delkey = NULL;
+		markertmp = NULL;
 		for (node = PRIVKEY_LIST_HEAD(privkeys);
 		     !PRIVKEY_LIST_EMPTY(privkeys) &&
 		     !PRIVKEY_LIST_END(node, privkeys);
 		     node = PRIVKEY_LIST_NEXT(node)) {
 			keyname = PK11_GetPrivateKeyNickname(node->key);
 			if ((keyname != NULL) &&
-			    (entry->cm_key_nickname != NULL) &&
-			    (strcmp(keyname, entry->cm_key_nickname) == 0)) {
-				/* Avoid stealing the key reference from the
-				 * list. */
-				delkey = SECKEY_CopyPrivateKey(node->key);
+			    (strcmp(keyname, nickname) == 0)) {
+				/* We're going to need to use a different nickname. */
+				cm_log(1, "Key already exists with nickname \"%s\".\n", nickname);
+				free(nickname);
+				nickname = make_nickname(entry->cm_key_nickname, &markertmp);
 				break;
 			}
 		}
 		SECKEY_DestroyPrivateKeyList(privkeys);
-		if (delkey != NULL) {
-			PK11_DeleteTokenPrivateKey(delkey, PR_TRUE);
-			/* If we found at least one key before, scan again. */
+		if (markertmp != NULL) {
+			/* If we found at least one match, scan again for the new nickname. */
 			privkeys = PK11_ListPrivKeysInSlot(slot,
-							   entry->cm_key_nickname,
+							   nickname,
 							   NULL);
+			marker = markertmp;
 		} else {
+			cm_log(1, "Nickname \"%s\" appears to be unused.\n", nickname);
 			privkeys = NULL;
 		}
 	}
+	if ((marker == NULL) || (strlen(marker) == 0)) {
+		/* Look harder.  Walk the list of certificates in the token,
+		 * looking at each one to see if it matches the specified
+		 * nickname. */
+		markertmp = NULL;
+		certs = PK11_ListCertsInSlot(slot);
+		while (certs != NULL) {
+			cert = NULL;
+			for (cnode = CERT_LIST_HEAD(certs);
+			     !CERT_LIST_EMPTY(certs) &&
+			     !CERT_LIST_END(cnode, certs);
+			     cnode = CERT_LIST_NEXT(cnode)) {
+				cert = cnode->cert;
+				if ((nickname != NULL) &&
+				    (strcmp(cert->nickname, nickname) == 0)) {
+					cm_log(3, "Located a certificate with "
+					       "the desired nickname (\"%s\").\n",
+					       nickname);
+					ckey = PK11_FindPrivateKeyFromCert(slot,
+									   cert,
+									   NULL);
+					if (ckey != NULL) {
+						cm_log(3, "And we found "
+						       "its private key.\n");
+						SECKEY_DestroyPrivateKey(ckey);
+					} else {
+						cm_log(3, "But we didn't find "
+						       "its private key.\n");
+					}
+					break;
+				}
+				cert = NULL;
+			}
+			if (cert == NULL) {
+				cm_log(1, "Nickname \"%s\" appears to be unused.\n", nickname);
+				CERT_DestroyCertList(certs);
+				certs = NULL;
+			} else {
+				free(nickname);
+				nickname = make_nickname(entry->cm_key_nickname, &markertmp);
+				marker = markertmp;
+			}
+		}
+	}
+
 	/* Attach the specified nickname to the key. */
-	error = PK11_SetPrivateKeyNickname(privkey, entry->cm_key_nickname);
+	error = PK11_SetPrivateKeyNickname(privkey, nickname);
 	if (error != SECSuccess) {
 		ec = PORT_GetError();
 		if (ec != 0) {
@@ -560,10 +670,11 @@ retry_gen:
 			es = NULL;
 		}
 		if (es != NULL) {
-			cm_log(1, "Error setting nickname on private key: "
-			       "%s.\n", es);
+			cm_log(1, "Error setting nickname \"%s\" on private key: "
+			       "%s.\n", nickname, es);
 		} else {
-			cm_log(1, "Error setting nickname on private key.\n");
+			cm_log(1, "Error setting nickname \"%s\" on private key.\n",
+			       nickname);
 		}
 		switch (ec) {
 		case PR_NO_ACCESS_RIGHTS_ERROR: /* EACCES or EPERM */
@@ -574,11 +685,12 @@ retry_gen:
 			break;
 		}
 	}
+	cm_log(1, "Set nickname \"%s\" on private key.\n", nickname);
 	/* Encode the public key to hex, and print it. */
 	spki = SECKEY_EncodeDERSubjectPublicKeyInfo(pubkey);
 	if (spki != NULL) {
 		pubihex = cm_store_hex_from_bin(NULL, spki->data,
-					        spki->len);
+						spki->len);
 		SECITEM_FreeItem(spki, PR_TRUE);
 	} else {
 		pubihex = "";
@@ -592,11 +704,47 @@ retry_gen:
 	} else {
 		pubhex = "";
 	}
-	fprintf(status, "%s\n%s\n", pubihex, pubhex);
+	fprintf(status, "%s\n%s\n%s\n", pubihex, pubhex, marker ? marker : "");
 	SECKEY_DestroyPrivateKey(privkey);
 	SECKEY_DestroyPublicKey(pubkey);
+	/* Try to remove any keys with old candidate names. */
+	if ((entry->cm_key_next_marker != NULL) &&
+	    (strlen(entry->cm_key_next_marker) > 0)) {
+		free(nickname);
+		nickname = util_build_next_nickname(entry->cm_key_nickname, entry->cm_key_next_marker);
+		privkeys = PK11_ListPrivKeysInSlot(slot, nickname, NULL);
+		while ((privkeys != NULL) && !PRIVKEY_LIST_EMPTY(privkeys)) {
+			delkey = NULL;
+			for (node = PRIVKEY_LIST_HEAD(privkeys);
+			     !PRIVKEY_LIST_EMPTY(privkeys) &&
+			     !PRIVKEY_LIST_END(node, privkeys);
+			     node = PRIVKEY_LIST_NEXT(node)) {
+				keyname = PK11_GetPrivateKeyNickname(node->key);
+				if ((keyname != NULL) && (nickname != NULL) &&
+				    (strcmp(keyname, nickname) == 0)) {
+					/* Avoid stealing the key reference from the
+					 * list. */
+					delkey = SECKEY_CopyPrivateKey(node->key);
+					break;
+				}
+			}
+			SECKEY_DestroyPrivateKeyList(privkeys);
+			if (delkey != NULL) {
+				PK11_DeleteTokenPrivateKey(delkey, PR_FALSE);
+				cm_log(1, "Removing key with nickname \"%s\".\n", nickname);
+				/* If we found at least one key before, scan again. */
+				privkeys = PK11_ListPrivKeysInSlot(slot,
+								   nickname,
+								   NULL);
+			} else {
+				privkeys = NULL;
+			}
+		}
+	}
 	PK11_FreeSlotList(slotlist);
 	error = NSS_ShutdownContext(ctx);
+	free(nickname);
+	util_set_db_entry_key_owner(entry->cm_key_storage_location, entry);
 	if (error != SECSuccess) {
 		cm_log(1, "Error shutting down NSS.\n");
 	}
@@ -606,25 +754,24 @@ retry_gen:
 
 /* Check if the keypair is ready. */
 static int
-cm_keygen_n_ready(struct cm_store_entry *entry, struct cm_keygen_state *state)
+cm_keygen_n_ready(struct cm_keygen_state *state)
 {
-	return cm_subproc_ready(entry, state->subproc);
+	return cm_subproc_ready(state->subproc);
 }
 
 /* Get a selectable-for-read descriptor we can poll for status changes. */
 static int
-cm_keygen_n_get_fd(struct cm_store_entry *entry, struct cm_keygen_state *state)
+cm_keygen_n_get_fd(struct cm_keygen_state *state)
 {
-	return cm_subproc_get_fd(entry, state->subproc);
+	return cm_subproc_get_fd(state->subproc);
 }
 
 /* Tell us if the keypair was saved to the location specified in the entry. */
 static int
-cm_keygen_n_saved_keypair(struct cm_store_entry *entry,
-		          struct cm_keygen_state *state)
+cm_keygen_n_saved_keypair(struct cm_keygen_state *state)
 {
 	int status;
-	status = cm_subproc_get_exitstatus(entry, state->subproc);
+	status = cm_subproc_get_exitstatus(state->subproc);
 	if (WIFEXITED(status) && (WEXITSTATUS(status) == 0)) {
 		return 0;
 	}
@@ -633,11 +780,10 @@ cm_keygen_n_saved_keypair(struct cm_store_entry *entry,
 
 /* Tell us if we don't have permissions. */
 static int
-cm_keygen_n_need_perms(struct cm_store_entry *entry,
-		       struct cm_keygen_state *state)
+cm_keygen_n_need_perms(struct cm_keygen_state *state)
 {
 	int status;
-	status = cm_subproc_get_exitstatus(entry, state->subproc);
+	status = cm_subproc_get_exitstatus(state->subproc);
 	if (WIFEXITED(status) &&
 	    (WEXITSTATUS(status) == CM_SUB_STATUS_ERROR_PERMS)) {
 		return 0;
@@ -647,11 +793,10 @@ cm_keygen_n_need_perms(struct cm_store_entry *entry,
 
 /* Tell us if we need a new/correct PIN to use the key store. */
 static int
-cm_keygen_n_need_pin(struct cm_store_entry *entry,
-		     struct cm_keygen_state *state)
+cm_keygen_n_need_pin(struct cm_keygen_state *state)
 {
 	int status;
-	status = cm_subproc_get_exitstatus(entry, state->subproc);
+	status = cm_subproc_get_exitstatus(state->subproc);
 	if (WIFEXITED(status) &&
 	    (WEXITSTATUS(status) == CM_SUB_STATUS_ERROR_AUTH)) {
 		return 0;
@@ -661,11 +806,10 @@ cm_keygen_n_need_pin(struct cm_store_entry *entry,
 
 /* Check if we need a token to be inserted to generate the key. */
 static int
-cm_keygen_n_need_token(struct cm_store_entry *entry,
-		       struct cm_keygen_state *state)
+cm_keygen_n_need_token(struct cm_keygen_state *state)
 {
 	int status;
-	status = cm_subproc_get_exitstatus(entry, state->subproc);
+	status = cm_subproc_get_exitstatus(state->subproc);
 	if (WIFEXITED(status) &&
 	    (WEXITSTATUS(status) == CM_SUB_STATUS_ERROR_NO_TOKEN)) {
 		return 0;
@@ -675,28 +819,47 @@ cm_keygen_n_need_token(struct cm_store_entry *entry,
 
 /* Clean up after key generation. */
 static void
-cm_keygen_n_done(struct cm_store_entry *entry, struct cm_keygen_state *state)
+cm_keygen_n_done(struct cm_keygen_state *state)
 {
-	const char *pubkey_info, *p;
+	const char *output, *p, *q;
+	char *pubkey_info, *pubkey, *marker = NULL;
 	int len;
 
 	if (state->subproc != NULL) {
-		pubkey_info = cm_subproc_get_msg(entry, state->subproc, NULL);
-		if (pubkey_info != NULL) {
-			len = strcspn(pubkey_info, "\r\n");
-			entry->cm_key_pubkey_info = talloc_strndup(entry,
-								   pubkey_info,
-								   len);
-			p = pubkey_info + len;
-			p += strspn(p, "\r\n");
+		output = cm_subproc_get_msg(state->subproc, NULL);
+		if (output != NULL) {
+			p = output;
+			len = strcspn(output, "\r\n");
+			pubkey_info = talloc_strndup(state->entry, p, len);
+			q = p + len;
+			p = q + strspn(q, "\r\n");
 			len = strcspn(p, "\r\n");
-			entry->cm_key_pubkey = talloc_strndup(entry,
-							      p, len);
-		} else {
-			entry->cm_key_pubkey_info = NULL;
-			entry->cm_key_pubkey = NULL;
+			pubkey = talloc_strndup(state->entry, p, len);
+			q = p + len;
+			p = q + strspn(q, "\r\n");
+			len = strcspn(p, "\r\n");
+			if (len > 0) {
+				marker = talloc_strndup(state->entry, p, len);
+			}
+			if ((marker != NULL) && (strlen(marker) > 0)) {
+				state->entry->cm_key_next_pubkey_info = pubkey_info;
+				state->entry->cm_key_next_pubkey = pubkey;
+				state->entry->cm_key_next_marker = marker;
+				state->entry->cm_key_next_generated_date = time(NULL);
+				state->entry->cm_key_next_requested_count = 0;
+			} else {
+				state->entry->cm_key_next_pubkey_info = NULL;
+				state->entry->cm_key_next_pubkey = NULL;
+				state->entry->cm_key_next_marker = NULL;
+				state->entry->cm_key_next_generated_date = 0;
+				state->entry->cm_key_pubkey_info = pubkey_info;
+				state->entry->cm_key_pubkey = pubkey;
+				state->entry->cm_key_generated_date = time(NULL);
+				state->entry->cm_key_requested_count = 0;
+				state->entry->cm_key_issued_count = 0;
+			}
 		}
-		cm_subproc_done(entry, state->subproc);
+		cm_subproc_done(state->subproc);
 	}
 	talloc_free(state);
 }
@@ -722,7 +885,8 @@ cm_keygen_n_start(struct cm_store_entry *entry)
 		state->pvt.need_pin = cm_keygen_n_need_pin;
 		state->pvt.need_token = cm_keygen_n_need_token;
 		state->pvt.done = cm_keygen_n_done;
-		state->subproc = cm_subproc_start(cm_keygen_n_main,
+		state->entry = entry;
+		state->subproc = cm_subproc_start(cm_keygen_n_main, state,
 						  NULL, entry, &settings);
 		if (state->subproc == NULL) {
 			talloc_free(state);
