@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2009,2010,2011,2012,2014 Red Hat, Inc.
+ * Copyright (C) 2009,2010,2011,2012,2014,2015 Red Hat, Inc.
  * 
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -57,13 +57,13 @@ cm_keyiread_o_main(int fd, struct cm_store_ca *ca, struct cm_store_entry *entry,
 {
 	struct cm_pin_cb_data cb_data;
 	FILE *pem, *fp;
-	EVP_PKEY *pkey;
+	EVP_PKEY *pkey, *nextpkey = NULL;
 	int status;
 	char buf[LINE_MAX];
 	const char *alg;
 	int bits, length;
 	long error;
-	char *pin, *pubkey, *pubikey;
+	char *pin, *pubkey, *pubikey, *nextfile;
 	unsigned char *tmp;
 
 	util_o_init();
@@ -104,6 +104,39 @@ cm_keyiread_o_main(int fd, struct cm_store_ca *ca, struct cm_store_entry *entry,
 			}
 		}
 		fclose(pem);
+		if ((status == 0) &&
+		    (entry->cm_key_next_marker != NULL) &&
+		    (strlen(entry->cm_key_next_marker) > 0)) {
+			nextfile = util_build_next_filename(entry->cm_key_storage_location, entry->cm_key_next_marker);
+			pem = fopen(nextfile, "r");
+			if (pem != NULL) {
+				nextpkey = PEM_read_PrivateKey(pem, NULL,
+							       cm_pin_read_for_key_ossl_cb,
+							       &cb_data);
+				if (nextpkey == NULL) {
+					cm_log(1, "Internal error reading key from \"%s\".\n",
+					       nextfile);
+					status = CM_SUB_STATUS_ERROR_AUTH; /* XXX */
+				} else {
+					if ((pin != NULL) &&
+					    (strlen(pin) > 0) &&
+					    (cb_data.n_attempts == 0)) {
+						cm_log(1, "PIN was not needed to read private "
+						       "key '%s', though one was provided. "
+						       "Treating this as an error.\n",
+						       nextfile);
+						status = CM_SUB_STATUS_ERROR_AUTH; /* XXX */
+					}
+				}
+				fclose(pem);
+			} else {
+				cm_log(1, "Error opening key file '%s' "
+				       "for reading: %s.\n",
+				       nextfile, strerror(errno));
+				nextpkey = NULL;
+			}
+			free(nextfile);
+		}
 	} else {
 		if (errno != ENOENT) {
 			cm_log(1, "Error opening key file '%s' "
@@ -155,6 +188,45 @@ cm_keyiread_o_main(int fd, struct cm_store_ca *ca, struct cm_store_entry *entry,
 			}
 		}
 		fprintf(fp, "%s/%d/%s/%s\n", alg, bits, pubikey, pubkey);
+		if (nextpkey != NULL) {
+			switch (EVP_PKEY_type(nextpkey->type)) {
+			case EVP_PKEY_RSA:
+				cm_log(3, "Next key is an RSA key.\n");
+				alg = "RSA";
+				break;
+#ifdef CM_ENABLE_DSA
+			case EVP_PKEY_DSA:
+				cm_log(3, "Next key is a DSA key.\n");
+				alg = "DSA";
+				break;
+#endif
+#ifdef CM_ENABLE_EC
+			case EVP_PKEY_EC:
+				cm_log(3, "Next key is an EC key.\n");
+				alg = "EC";
+				break;
+#endif
+			default:
+				cm_log(3, "Next key is for an unknown algorithm.\n");
+				alg = "";
+				break;
+			}
+			bits = EVP_PKEY_bits(nextpkey);
+			cm_log(3, "Next key size is %d.\n", bits);
+			tmp = NULL;
+			length = i2d_PUBKEY(nextpkey, (unsigned char **) &tmp);
+			if (length > 0) {
+				pubikey = cm_store_hex_from_bin(NULL, tmp, length);
+			}
+			tmp = NULL;
+			length = i2d_PublicKey(nextpkey, (unsigned char **) &tmp);
+			if (length > 0) {
+				pubkey = cm_store_hex_from_bin(NULL, tmp, length);
+			}
+			fprintf(fp, "%s/%d/%s/%s\n", alg, bits, pubikey, pubkey);
+		} else {
+			fprintf(fp, "\n");
+		}
 		status = 0;
 	} else {
 		while ((error = ERR_get_error()) != 0) {
