@@ -25,6 +25,7 @@
 #include <stdlib.h>
 #include <string.h>
 #include <time.h>
+#include <unistd.h>
 
 #include <talloc.h>
 
@@ -36,11 +37,10 @@
 			"0123456789" \
 			"+/="
 
-static struct {
+static const struct {
 	const char *name;
 	enum cm_state state;
 } cm_state_names[] = {
-	{"INVALID", CM_INVALID},
 	{"NEED_KEY_PAIR", CM_NEED_KEY_PAIR},
 	{"GENERATING_KEY_PAIR", CM_GENERATING_KEY_PAIR},
 	{"NEED_KEY_GEN_PERMS", CM_NEED_KEY_GEN_PERMS},
@@ -57,6 +57,13 @@ static struct {
 	{"NEED_CSR_GEN_PIN", CM_NEED_CSR_GEN_PIN},
 	{"NEED_CSR_GEN_TOKEN", CM_NEED_CSR_GEN_TOKEN},
 	{"HAVE_CSR", CM_HAVE_CSR},
+	{"NEED_SCEP_DATA", CM_NEED_SCEP_DATA},
+	{"GENERATING_SCEP_DATA", CM_GENERATING_SCEP_DATA},
+	{"NEED_SCEP_GEN_PIN", CM_NEED_SCEP_GEN_PIN},
+	{"NEED_SCEP_GEN_TOKEN", CM_NEED_SCEP_GEN_TOKEN},
+	{"NEED_SCEP_ENCRYPTION_CERT", CM_NEED_SCEP_ENCRYPTION_CERT},
+	{"NEED_SCEP_RSA_CLIENT_KEY", CM_NEED_SCEP_RSA_CLIENT_KEY},
+	{"HAVE_SCEP_DATA", CM_HAVE_SCEP_DATA},
 	{"NEED_TO_SUBMIT", CM_NEED_TO_SUBMIT},
 	{"SUBMITTING", CM_SUBMITTING},
 	{"NEED_CA", CM_NEED_CA},
@@ -69,6 +76,8 @@ static struct {
 	{"START_SAVING_CERT", CM_START_SAVING_CERT},
 	{"SAVING_CERT", CM_SAVING_CERT},
 	{"NEED_CERTSAVE_PERMS", CM_NEED_CERTSAVE_PERMS},
+	{"NEED_CERTSAVE_TOKEN", CM_NEED_CERTSAVE_TOKEN},
+	{"NEED_CERTSAVE_PIN", CM_NEED_CERTSAVE_PIN},
 	{"NEED_TO_READ_CERT", CM_NEED_TO_READ_CERT},
 	{"READING_CERT", CM_READING_CERT},
 	{"SAVED_CERT", CM_SAVED_CERT},
@@ -78,8 +87,14 @@ static struct {
 	{"NOTIFYING_VALIDITY", CM_NOTIFYING_VALIDITY},
 	{"NEED_TO_NOTIFY_REJECTION", CM_NEED_TO_NOTIFY_REJECTION},
 	{"NOTIFYING_REJECTION", CM_NOTIFYING_REJECTION},
-	{"NEED_TO_NOTIFY_ISSUED_FAILED", CM_NEED_TO_NOTIFY_ISSUED_FAILED},
-	{"NOTIFYING_ISSUED_FAILED", CM_NOTIFYING_ISSUED_FAILED},
+	{"NEED_TO_NOTIFY_ISSUED_SAVE_FAILED", CM_NEED_TO_NOTIFY_ISSUED_SAVE_FAILED},
+	{"NOTIFYING_ISSUED_SAVE_FAILED", CM_NOTIFYING_ISSUED_SAVE_FAILED},
+	{"NEED_TO_NOTIFY_ISSUED_CA_SAVE_FAILED", CM_NEED_TO_NOTIFY_ISSUED_CA_SAVE_FAILED},
+	{"NOTIFYING_ISSUED_CA_SAVE_FAILED", CM_NOTIFYING_ISSUED_CA_SAVE_FAILED},
+	{"NEED_TO_NOTIFY_ONLY_CA_SAVE_FAILED", CM_NEED_TO_NOTIFY_ONLY_CA_SAVE_FAILED},
+	{"NOTIFYING_ONLY_CA_SAVE_FAILED", CM_NOTIFYING_ONLY_CA_SAVE_FAILED},
+	{"NEED_TO_SAVE_CA_CERTS", CM_NEED_TO_SAVE_CA_CERTS},
+	{"NEED_TO_SAVE_ONLY_CA_CERTS", CM_NEED_TO_SAVE_ONLY_CA_CERTS},
 	{"NEED_TO_NOTIFY_ISSUED_SAVED", CM_NEED_TO_NOTIFY_ISSUED_SAVED},
 	{"NOTIFYING_ISSUED_SAVED", CM_NOTIFYING_ISSUED_SAVED},
 	{"NEED_GUIDANCE", CM_NEED_GUIDANCE},
@@ -91,14 +106,165 @@ static struct {
 	{"NEWLY_ADDED_START_READING_CERT", CM_NEWLY_ADDED_START_READING_CERT},
 	{"NEWLY_ADDED_READING_CERT", CM_NEWLY_ADDED_READING_CERT},
 	{"NEWLY_ADDED_DECIDING", CM_NEWLY_ADDED_DECIDING},
-	/* old names */
+	{"START_SAVING_CA_CERTS", CM_START_SAVING_CA_CERTS},
+	{"SAVING_CA_CERTS", CM_SAVING_CA_CERTS},
+	{"START_SAVING_ONLY_CA_CERTS", CM_START_SAVING_ONLY_CA_CERTS},
+	{"SAVING_ONLY_CA_CERTS", CM_SAVING_ONLY_CA_CERTS},
+	{"NEED_CA_CERT_SAVE_PERMS", CM_NEED_CA_CERT_SAVE_PERMS},
+	{"NEED_ONLY_CA_CERT_SAVE_PERMS", CM_NEED_ONLY_CA_CERT_SAVE_PERMS},
+	{"INVALID", CM_INVALID},
+	/* old names for since-renamed states */
 	{"NEED_TO_NOTIFY", CM_NEED_TO_NOTIFY_VALIDITY},
 	{"NOTIFYING", CM_NOTIFYING_VALIDITY},
 	{"NEWLY_ADDED_START_READING_KEYI", CM_NEWLY_ADDED_START_READING_KEYINFO},
 	{"NEWLY_ADDED_READING_KEYI", CM_NEWLY_ADDED_READING_KEYINFO},
 	{"NEWLY_ADDED_NEED_KEYI_READ_PIN", CM_NEWLY_ADDED_NEED_KEYINFO_READ_PIN},
+	{"NEED_TO_NOTIFY_ISSUED_FAILED", CM_NEED_TO_NOTIFY_ISSUED_SAVE_FAILED},
+	{"NOTIFYING_ISSUED_FAILED", CM_NOTIFYING_ISSUED_SAVE_FAILED},
 };
 
+static const struct {
+	const char *name;
+	enum cm_ca_phase_state state;
+} cm_ca_state_names[] = {
+	{"IDLE", CM_CA_IDLE},
+	{"NEED_TO_REFRESH", CM_CA_NEED_TO_REFRESH},
+	{"REFRESHING", CM_CA_REFRESHING},
+	{"UNREACHABLE", CM_CA_DATA_UNREACHABLE},
+	{"NEED_TO_SAVE_DATA", CM_CA_NEED_TO_SAVE_DATA},
+	{"PRE_SAVE_DATA", CM_CA_PRE_SAVE_DATA},
+	{"START_SAVING_DATA", CM_CA_START_SAVING_DATA},
+	{"SAVING_DATA", CM_CA_SAVING_DATA,},
+	{"NEED_POST_SAVE_DATA", CM_CA_NEED_POST_SAVE_DATA},
+	{"POST_SAVE_DATA", CM_CA_POST_SAVE_DATA},
+	{"SAVED_DATA", CM_CA_SAVED_DATA},
+	{"NEED_TO_ANALYZE", CM_CA_NEED_TO_ANALYZE},
+	{"ANALYZING", CM_CA_ANALYZING},
+	{"DISABLED", CM_CA_DISABLED},
+};
+
+static const struct {
+	const char *name;
+	enum cm_ca_phase phase;
+} cm_ca_phase_names[] = {
+	{"identify", cm_ca_phase_identify},
+	{"certs", cm_ca_phase_certs},
+	{"profiles", cm_ca_phase_profiles},
+	{"default_profile", cm_ca_phase_default_profile},
+	{"enrollment_reqs", cm_ca_phase_enroll_reqs},
+	{"renewal_reqs", cm_ca_phase_renew_reqs},
+	{"capabilities", cm_ca_phase_capabilities},
+	{"encryption_certs", cm_ca_phase_encryption_certs},
+	{"invalid", cm_ca_phase_invalid},
+};
+
+const char *
+cm_store_ca_state_as_string(enum cm_ca_phase_state state)
+{
+	unsigned int i;
+
+	for (i = 0;
+	     i < sizeof(cm_ca_state_names) / sizeof(cm_ca_state_names[0]);
+	     i++) {
+		if (cm_ca_state_names[i].state == state) {
+			return cm_ca_state_names[i].name;
+		}
+	}
+	return "UNKNOWN";
+}
+
+const char *
+cm_store_ca_phase_as_string(enum cm_ca_phase phase)
+{
+	unsigned int i;
+
+	for (i = 0;
+	     i < sizeof(cm_ca_phase_names) / sizeof(cm_ca_phase_names[0]);
+	     i++) {
+		if (cm_ca_phase_names[i].phase == phase) {
+			return cm_ca_phase_names[i].name;
+		}
+	}
+	return "invalid";
+}
+
+const char *
+cm_store_state_as_string(enum cm_state state)
+{
+	unsigned int i;
+	for (i = 0;
+	     i < sizeof(cm_state_names) / sizeof(cm_state_names[0]);
+	     i++) {
+		if (cm_state_names[i].state == state) {
+			return cm_state_names[i].name;
+		}
+	}
+	return "UNKNOWN";
+}
+
+enum cm_ca_phase_state
+cm_store_ca_state_from_string(const char *name)
+{
+	unsigned long l;
+	unsigned i;
+	char *p;
+
+	for (i = 0;
+	     i < sizeof(cm_ca_state_names) / sizeof(cm_ca_state_names[0]);
+	     i++) {
+		if (strcasecmp(cm_ca_state_names[i].name, name) == 0) {
+			return cm_ca_state_names[i].state;
+		}
+	}
+	l = strtoul(name, &p, 10);
+	if ((*name != '\0') && (p != NULL) && (*p == '\0')) {
+		return l;
+	}
+	return CM_CA_DISABLED;
+}
+
+enum cm_ca_phase
+cm_store_ca_phase_from_string(const char *name)
+{
+	unsigned long l;
+	unsigned int i;
+	char *p;
+
+	for (i = 0;
+	     i < sizeof(cm_ca_phase_names) / sizeof(cm_ca_phase_names[0]);
+	     i++) {
+		if (strcasecmp(cm_ca_phase_names[i].name, name) == 0) {
+			return cm_ca_phase_names[i].phase;
+		}
+	}
+	l = strtoul(name, &p, 10);
+	if ((*name != '\0') && (p != NULL) && (*p == '\0')) {
+		return l;
+	}
+	return cm_ca_phase_invalid;
+}
+
+enum cm_state
+cm_store_state_from_string(const char *name)
+{
+	unsigned long l;
+	unsigned int i;
+	char *p;
+
+	for (i = 0;
+	     i < sizeof(cm_state_names) / sizeof(cm_state_names[0]);
+	     i++) {
+		if (strcasecmp(cm_state_names[i].name, name) == 0) {
+			return cm_state_names[i].state;
+		}
+	}
+	l = strtoul(name, &p, 10);
+	if ((*name != '\0') && (p != NULL) && (*p == '\0')) {
+		return l;
+	}
+	return CM_INVALID;
+}
+
 char *
 cm_store_maybe_strdup(void *parent, const char *s)
 {
@@ -128,34 +294,6 @@ cm_store_maybe_strdupv(void *parent, char **s)
 	return ret;
 }
 
-const char *
-cm_store_state_as_string(enum cm_state state)
-{
-	unsigned int i;
-	for (i = 0;
-	     i < sizeof(cm_state_names) / sizeof(cm_state_names[0]);
-	     i++) {
-		if (cm_state_names[i].state == state) {
-			return cm_state_names[i].name;
-		}
-	}
-	return "UNKNOWN";
-}
-
-enum cm_state
-cm_store_state_from_string(const char *name)
-{
-	unsigned int i;
-	for (i = 0;
-	     i < sizeof(cm_state_names) / sizeof(cm_state_names[0]);
-	     i++) {
-		if (strcasecmp(cm_state_names[i].name, name) == 0) {
-			return cm_state_names[i].state;
-		}
-	}
-	return CM_INVALID;
-}
-
 /* Generic routines. */
 struct cm_store_entry *
 cm_store_entry_new(void *parent)
@@ -398,10 +536,29 @@ cm_store_hex_to_bin(const char *serial, unsigned char *buf, int length)
 }
 
 char *
-cm_store_canonicalize_directory(void *parent, const char *path)
+cm_store_canonicalize_path(void *parent, const char *path)
 {
-	char *tmp, *p;
+	char *tmp = NULL, *p, *q, buf[PATH_MAX], *prefix;
 	int i;
+
+	if (strncmp(path, "dbm:", 4) == 0) {
+		prefix = "dbm";
+		path += 4;
+	} else
+	if (strncmp(path, "sql:", 4) == 0) {
+		prefix = "sql";
+		path += 4;
+	} else
+	if (strncmp(path, "rdb:", 4) == 0) {
+		prefix = "rdb";
+		path += 4;
+	} else
+	if (strncmp(path, "extern:", 4) == 0) {
+		prefix = "extern";
+		path += 7;
+	} else {
+		prefix = NULL;
+	}
 	i = strlen(path);
 	if (i > 1) {
 		while ((i > 1) && (path[i - 1] == '/')) {
@@ -411,11 +568,43 @@ cm_store_canonicalize_directory(void *parent, const char *path)
 	} else {
 		tmp = talloc_strdup(parent, path);
 	}
-	while ((p = strstr(tmp, "/./")) != NULL) {
-		memmove(p, p + 2, strlen(p) - 1);
+	if ((tmp != NULL) && (tmp[0] != '/')) {
+		memset(buf, '\0', sizeof(buf));
+		if (getcwd(buf, sizeof(buf) - 1) != NULL) {
+			tmp = talloc_asprintf(parent, "%s//%s", buf, tmp);
+		}
 	}
-	while ((p = strstr(tmp, "//")) != NULL) {
-		memmove(p, p + 1, strlen(p));
+	if (tmp != NULL) {
+		for (p = tmp; *p != '\0'; p++) {
+			if ((strncmp(p, "/", 1) == 0) &&
+			    ((p[1] == '/') || (p[1] == '\0'))) {
+				memmove(p, p + 1, strlen(p + 1) + 1);
+			}
+		}
+		for (p = tmp; *p != '\0'; p++) {
+			if ((strncmp(p, "/.", 2) == 0) &&
+			    ((p[2] == '/') || (p[2] == '\0'))) {
+				q = p - 1;
+				memmove(p, p + 2, strlen(p + 2) + 1);
+			}
+		}
+		for (p = tmp; *p != '\0'; p++) {
+			if ((strncmp(p, "/..", 3) == 0) &&
+			    ((p[3] == '/') || (p[3] == '\0'))) {
+				q = p - 1;
+				while ((q >= tmp) && (*q != '/')) {
+					q--;
+				}
+				if (*q == '/') {
+					memmove(q, p + 3, strlen(p + 3) + 1);
+				} else {
+					break;
+				}
+			}
+		}
+		if (prefix != NULL) {
+			tmp = talloc_asprintf(parent, "%s:%s", prefix, tmp);
+		}
 	}
 	return tmp;
 }
@@ -548,7 +737,7 @@ cm_store_utf8_from_bmp_string(unsigned char *bmp, unsigned int len)
 }
 
 char *
-cm_store_base64_from_bin(void *parent, unsigned char *buf, int length)
+cm_store_base64_from_bin(void *parent, const unsigned char *buf, int length)
 {
 	char *p, *ret;
 	int max, i, j;
@@ -608,8 +797,6 @@ cm_store_base64_to_bin(const char *serial, int insize,
 	unsigned char *b;
 	uint32_t u, count;
 
-	p = serial;
-	b = buf;
 	u = 0;
 	count = 0;
 	if (insize < 0) {
