@@ -51,7 +51,7 @@ char *cm_store_hex_from_bin(void *parent,
 			    const unsigned char *serial, int length);
 int cm_store_hex_to_bin(const char *serial, unsigned char *buf, int length);
 char *cm_store_base64_from_bin(void *parent,
-			       unsigned char *buf, int length);
+			       const unsigned char *buf, int length);
 int cm_store_base64_to_bin(const char *serial, int insize,
 			   unsigned char *buf, int maxlength);
 char *cm_store_base64_as_bin(void *parent, const char *serial, int insize,
