@@ -30,6 +30,8 @@ enum cm_external_status {
 	CM_SUBMIT_STATUS_UNCONFIGURED = 4,
 	CM_SUBMIT_STATUS_WAIT_WITH_DELAY = 5,
 	CM_SUBMIT_STATUS_OPERATION_NOT_SUPPORTED = 6,
+	CM_SUBMIT_STATUS_NEED_SCEP_MESSAGES = 16,
+	CM_SUBMIT_STATUS_NEED_REKEY = 17,
 };
 const char *cm_submit_e_status_text(enum cm_external_status status);
 
@@ -37,6 +39,7 @@ const char *cm_submit_e_status_text(enum cm_external_status status);
 #define CM_SUBMIT_REQ_HOSTNAME_ENV "CERTMONGER_REQ_HOSTNAME"
 #define CM_SUBMIT_REQ_PRINCIPAL_ENV "CERTMONGER_REQ_PRINCIPAL"
 #define CM_SUBMIT_REQ_EMAIL_ENV "CERTMONGER_REQ_EMAIL"
+#define CM_SUBMIT_REQ_IP_ADDRESS_ENV "CERTMONGER_REQ_IP_ADDRESS"
 #define CM_SUBMIT_OPERATION_ENV "CERTMONGER_OPERATION"
 #define CM_SUBMIT_CSR_ENV "CERTMONGER_CSR"
 #define CM_SUBMIT_SPKAC_ENV "CERTMONGER_SPKAC"
@@ -46,6 +49,14 @@ const char *cm_submit_e_status_text(enum cm_external_status status);
 #define CM_SUBMIT_CA_NICKNAME_ENV "CERTMONGER_CA_NICKNAME"
 #define CM_SUBMIT_PROFILE_ENV "CERTMONGER_CA_PROFILE"
 #define CM_SUBMIT_CERTIFICATE_ENV "CERTMONGER_CERTIFICATE"
+#define CM_SUBMIT_SCEP_CA_IDENTIFIER_ENV "CERTMONGER_SCEP_CA_IDENTIFIER"
+#define CM_SUBMIT_SCEP_RA_CERTIFICATE_ENV "CERTMONGER_SCEP_RA_CERTIFICATE"
+#define CM_SUBMIT_SCEP_CA_CERTIFICATE_ENV "CERTMONGER_SCEP_CA_CERTIFICATE"
+#define CM_SUBMIT_SCEP_CERTIFICATES_ENV "CERTMONGER_SCEP_CERTIFICATES"
+#define CM_SUBMIT_SCEP_PKCSREQ_ENV "CERTMONGER_PKCSREQ"
+#define CM_SUBMIT_SCEP_PKCSREQ_REKEY_ENV "CERTMONGER_PKCSREQ_REKEY"
+#define CM_SUBMIT_SCEP_GETCERTINITIAL_ENV "CERTMONGER_GETCERTINITIAL"
+#define CM_SUBMIT_SCEP_GETCERTINITIAL_REKEY_ENV "CERTMONGER_GETCERTINITIAL_REKEY"
 
 #define CM_OP_SUBMIT "SUBMIT"
 #define CM_OP_POLL "POLL"
@@ -55,5 +66,7 @@ const char *cm_submit_e_status_text(enum cm_external_status status);
 #define CM_OP_FETCH_DEFAULT_PROFILE "GET-DEFAULT-TEMPLATE"
 #define CM_OP_FETCH_ENROLL_REQUIREMENTS "GET-NEW-REQUEST-REQUIREMENTS"
 #define CM_OP_FETCH_RENEWAL_REQUIREMENTS "GET-RENEW-REQUEST-REQUIREMENTS"
+#define CM_OP_FETCH_SCEP_CA_CERTS "FETCH-SCEP-CA-CERTS"
+#define CM_OP_FETCH_SCEP_CA_CAPS "FETCH-SCEP-CA-CAPS"
 
 #endif
