@@ -59,6 +59,11 @@ struct cm_subproc_state {
 	int fd, count, bufsize, status;
 };
 
+/* Start the passed callback in a subprocess, with a pipe that it can use to
+ * send data back to us.  If the callback exits, it must do so by calling
+ * _exit() or exec(), to avoid calling exit handlers registered by libraries
+ * that we use, which will screw us up.  Pretty much every bit of work that we
+ * can't do quickly is done this way. */
 struct cm_subproc_state *
 cm_subproc_start(int (*cb)(int fd,
 			   struct cm_store_ca *ca,
@@ -72,7 +77,7 @@ cm_subproc_start(int (*cb)(int fd,
 	struct cm_subproc_state *state;
 	int fds[2];
 	long flags;
-	char *configdir, *tmpdir, *tmp, *homedir, *bus, *local;
+	char *configdir, *tmpdir, *tmp, *homedir, *bus, *local, *pvt;
 
 	state = talloc_ptrtype(parent, state);
 	if (state != NULL) {
@@ -105,6 +110,8 @@ cm_subproc_start(int (*cb)(int fd,
 				bus = bus ? strdup(bus) : NULL;
 				local = cm_env_local_ca_dir();
 				local = local ? strdup(local) : NULL;
+				pvt = getenv(CERTMONGER_PVT_ADDRESS_ENV);
+				pvt = pvt ? strdup(pvt) : NULL;
 				clear_environment();
 				setenv("HOME", homedir, 1);
 				setenv("PATH", _PATH_STDPATH, 1);
@@ -121,12 +128,16 @@ cm_subproc_start(int (*cb)(int fd,
 					setenv("DBUS_SESSION_BUS_ADDRESS", bus,
 					       1);
 				}
+				if (pvt != NULL) {
+					setenv(CERTMONGER_PVT_ADDRESS_ENV, pvt,
+					       1);
+				}
 				if (local != NULL) {
 					setenv(CM_STORE_LOCAL_CA_DIRECTORY_ENV,
 					       local, 1);
 				}
 
-				exit((*cb)(fds[1], ca, entry, data));
+				_exit((*cb)(fds[1], ca, entry, data));
 				break;
 			default:
 				state->fd = fds[0];
@@ -377,13 +388,15 @@ cm_subproc_parse_args(void *parent, const char *cmdline, const char **error)
 }
 
 /* Redirect stdio to /dev/null, and mark everything else as close-on-exec,
- * except for perhaps one of them that is passed in by number. */
+ * except for perhaps one to three of them that are passed in by number. */
 void
-cm_subproc_mark_most_cloexec(int fd)
+cm_subproc_mark_most_cloexec(int fd, int fd2, int fd3)
 {
 	int i;
 	long l;
-	if (fd != STDIN_FILENO) {
+	if ((fd != STDIN_FILENO) &&
+	    (fd2 != STDIN_FILENO) &&
+	    (fd3 != STDIN_FILENO)) {
 		i = open("/dev/null", O_RDONLY);
 		if (i != -1) {
 			if (i != STDIN_FILENO) {
@@ -394,7 +407,9 @@ cm_subproc_mark_most_cloexec(int fd)
 			close(STDIN_FILENO);
 		}
 	}
-	if (fd != STDOUT_FILENO) {
+	if ((fd != STDOUT_FILENO) &&
+	    (fd2 != STDOUT_FILENO) &&
+	    (fd3 != STDOUT_FILENO)) {
 		i = open("/dev/null", O_WRONLY);
 		if (i != -1) {
 			if (i != STDOUT_FILENO) {
@@ -405,7 +420,9 @@ cm_subproc_mark_most_cloexec(int fd)
 			close(STDOUT_FILENO);
 		}
 	}
-	if (fd != STDERR_FILENO) {
+	if ((fd != STDERR_FILENO) &&
+	    (fd2 != STDERR_FILENO) &&
+	    (fd3 != STDERR_FILENO)) {
 		i = open("/dev/null", O_WRONLY);
 		if (i != -1) {
 			if (i != STDERR_FILENO) {
@@ -417,7 +434,9 @@ cm_subproc_mark_most_cloexec(int fd)
 		}
 	}
 	for (i = getdtablesize() - 1; i >= 3; i--) {
-		if (i == fd) {
+		if ((i == fd) ||
+		    (i == fd2) ||
+		    (i == fd3)) {
 			continue;
 		}
 		l = fcntl(i, F_GETFD);
