@@ -35,11 +35,18 @@ grep ^spkac= entry.nss.$size | sed s,spkac,SPKAC, > spkac.nss.$size
 # Generate a new CSR using the extracted key.
 $toolsdir/csrgen entry.openssl.$size > csr.openssl.$size
 grep ^spkac= entry.openssl.$size | sed s,spkac,SPKAC, > spkac.openssl.$size
+# Pull out the mini-cert.
+grep ^minicert= entry.openssl.$size | sed s,^minicert=,, | base64 -d > minicert.openssl.$size
+openssl x509 -out minicert.openssl.$size.pem -in minicert.openssl.$size -inform der
+grep ^minicert= entry.nss.$size | sed s,^minicert=,, | base64 -d > minicert.nss.$size
+openssl x509 -out minicert.nss.$size.pem -in minicert.nss.$size -inform der
 # The RSA tests already verify the contents of the requests, so we really only
 # need to care about the signatures passing verification.
 openssl req   -verify -noout < csr.nss.$size 2>&1
 openssl req   -verify -noout < csr.openssl.$size 2>&1
 openssl spkac -verify -noout < spkac.nss.$size 2>&1
 openssl spkac -verify -noout < spkac.openssl.$size 2>&1
+openssl verify -CAfile minicert.openssl.$size.pem minicert.openssl.$size.pem 2>&1
+openssl verify -CAfile minicert.nss.$size.pem minicert.nss.$size.pem 2>&1
 
 echo Test complete.
