@@ -28,6 +28,45 @@ Pi7xjf9BidMvYUEBpJsd9rA1LQtp/ZfxxA6RtgCeXjQPexjsvf6SLKyrmacHZcMJ
 b6JbhXMTzB7QZjR3IooqzXS8T/2zBxDUSH4fJ4o0KSkY8cjNCCxdnkXL96PC9KQ5
 kV1Ad3iHw/TnJjzrJJs3o92pRR/JtF0Jw6dszNP1Sn68uA==
 -----END CERTIFICATE-----"
+cert2="-----BEGIN CERTIFICATE-----
+MIIBoDCCAQmgAwIBAAIBATANBgkqhkiG9w0BAQsFADAUMRIwEAYDVQQDEwls
+b2NhbGhvc3QwIhgPMjAxNTAxMjcwMDAwMDBaGA8yMTE1MDEyNzAwMDAwMFow
+FDESMBAGA1UEAxMJbG9jYWxob3N0MIGfMA0GCSqGSIb3DQEBAQUAA4GNADCB
+iQKBgQDT7+BazrT3e/run8ZxrTfZNFx+vY+twKMPRXowSGot2eLcdDIIryOH
+ir1tUXIDZN2j+nF4U6kX3W66yMZjUApmYcFjhPk0Pg4ymsh/ScW2OlQXvC/f
+soPhvKA6cBeWUWwpdtRnFjZ14qmGuABPi6c/p0C/04HoBR9Y6QI5voRvHwID
+AQABMA0GCSqGSIb3DQEBCwUAA4GBAB4F1sjBaOJVuMmubbxc6vm3yDTwU3Qw
+JzjmXgwGUp5QryUIBZc9Kc5ceMUJ/Xf3OFDGWOqIx4JONdcgfLRJxax9WWg4
+mYbkAmUQBRtl7fGgEvOqF9EgtEY06Nj5aI7vbFEfB80Xd0O9O06ckxr7QBSc
+Wc2RCeFYrUpNi6s3vfM5
+-----END CERTIFICATE-----"
+cert3="-----BEGIN CERTIFICATE-----
+MIIB0TCCATqgAwIBAgICEjUwDQYJKoZIhvcNAQELBQAwFDESMBAGA1UEAxMJbG9j
+YWxob3N0MB4XDTcwMDEwMTExMTExMVoXDTcxMDEwMTExMTExMVowFDESMBAGA1UE
+AxMJbG9jYWxob3N0MIGfMA0GCSqGSIb3DQEBAQUAA4GNADCBiQKBgQC394oITlJc
+hLbXLUAYJ+IGpvr3YfyQlQb3u95p3s0gEemx/+Dy3ViJW7WhMuc/FHJSV5ghLflk
+TRLfUdS18LGBkfkREBJVWfELhgm5ZHqiGzCvTmg01tgvouKj+8kFNJIaWsEm8mBa
+SoJzk8+vOQsnXjrZAKAFY7WrVryilh4cQQIDAQABozIwMDAMBgNVHRMBAf8EAjAA
+MCAGA1UdDgEBAAQWBBQfjVS9H0rXGb4SFB9bkzTpHHUH2jANBgkqhkiG9w0BAQsF
+AAOBgQCt6xbyri3BobQUPQmN7ROc3mveMSfMyOwBSTDjl2XIWV98HjVLWRjScbg7
+KW6z8W7iaasSDF7GWM2YqWaanWx5XwzayNUvIX3gHKqo+OwHo5QUfawtEV5Niop1
+N4nZp1GMclRuk9UFXLV2NfaohRPYs7FVdBVtWNvg7hH8XedLyw==
+-----END CERTIFICATE-----"
+cert4="-----BEGIN CERTIFICATE-----
+MIIBoDCCAQmgAwIBAAIBATANBgkqhkiG9w0BAQsFADAUMRIwEAYDVQQDEwls
+b2NhbGhvc3QwIhgPMjAxNTAxMjcwMDAwMDBaGA8yMTE1MDEyNzAwMDAwMFow
+FDESMBAGA1UEAxMJbG9jYWxob3N0MIGfMA0GCSqGSIb3DQEBAQUAA4GNADCB
+iQKBgQDT7+BazrT3e/run8ZxrTfZNFx+vY+twKMPRXowSGot2eLcdDIIryOH
+ir1tUXIDZN2j+nF4U6kX3W66yMZjUApmYcFjhPk0Pg4ymsh/ScW2OlQXvC/f
+
+soPhvKA6cBeWUWwpdtRnFjZ14qmGuABPi6c/p0C/04HoBR9Y6QI5voRvHwID
+AQABMA0GCSqGSIb3DQEBCwUAA4GBAB4F1sjBaOJVuMmubbxc6vm3yDTwU3Qw
+JzjmXgwGUp5QryUIBZc9Kc5ceMUJ/Xf3OFDGWOqIx4JONdcgfLRJxax9WWg4
+mYbkAmUQBRtl7fGgEvOqF9EgtEY06Nj5aI7vbFEfB80Xd0O9O06ckxr7QBSc
+Wc2RCeFYrUpNi6s3vfM5
+
+-----END CERTIFICATE-----
+"
 
 cat > ca-issued << EOF
 #!/bin/sh
@@ -63,18 +102,51 @@ echo Also Monkeys
 exit 0
 EOF
 chmod u+x ca-issued-with-noise-both
+cat > ca-issued-json << EOF
+#!/bin/sh
+echo -n "{\"certificate\":\"$cert\",\"roots\":[{\"nickname\":\"CA1\",\"certificate\":\"$cert2\"},{\"nickname\":\"CA2\",\"certificate\":\"$cert3\"}],\"chain\":[{\"nickname\":\"CA3\",\"certificate\":\"$cert4\"}],\"whatever\":null}"
+exit 0
+EOF
+chmod u+x ca-issued-json
 cat > ca-ask-again << EOF
 #!/bin/sh
 echo iLoveCookiesSome
 exit 1
 EOF
 chmod u+x ca-ask-again
+cat > ca-need-rekey << EOF
+#!/bin/sh
+echo NotThatKeyAgain
+exit 17
+EOF
+chmod u+x ca-need-rekey
+cat > ca-issued-binary-x509 << EOF
+#!/bin/sh
+echo "$cert" | openssl x509 -outform der
+exit 0
+EOF
+chmod u+x ca-issued-binary-x509
 cat > ca-reject << EOF
 #!/bin/sh
 echo CA rejected us, must have been having a bad day.
 exit 2
 EOF
 chmod u+x ca-reject
+cat > ca-reject-second-time << EOF
+#!/bin/sh
+if test -z "\$CERTMONGER_CA_COOKIE" ; then
+	echo 1
+	echo Try again.
+	echo
+	echo Maybe later.
+	exit 5
+else
+	echo CA rejected us, must have been having a bad day.
+	echo cookie was "\$CERTMONGER_CA_COOKIE"
+	exit 2
+fi
+EOF
+chmod u+x ca-reject-second-time
 cat > ca-unreachable << EOF
 #!/bin/sh
 echo Could not contact CA.
@@ -107,6 +179,12 @@ echo What do you want?
 exit 6
 EOF
 chmod u+x ca-what-what-6
+cat > ca-needs-scep-16 << EOF
+#!/bin/sh
+echo Nope, need SCEP data.
+exit 16
+EOF
+chmod u+x ca-needs-scep-16
 
 cat > ca << EOF
 id=SelfSign
@@ -122,6 +200,10 @@ key_storage_location=$tmpdir/keyfile
 cert_storage_type=FILE
 cert_storage_location=$tmpdir/certfile
 notification_method=STDOUT
+post_certsave_command=echo POSTHOOK
+post_certsave_uid=`id -u`
+pre_certsave_command=echo PREHOOK
+pre_certsave_uid=`id -u`
 EOF
 # These cover parts of the process, forcing it to stop if any phase needs
 # to be tried again, so that we don't hit infinite loops.
@@ -132,6 +214,7 @@ if test "`grep ^state entry`" != state=NEED_KEYINFO ; then
 	grep ^state entry
 	exit 1
 fi
+grep ^key.\*count= entry | LANG=C sort
 
 echo
 echo '[Reading back key info.]'
@@ -142,6 +225,7 @@ if test "`grep ^state entry`" != state=NEED_CSR ; then
 	exit 1
 fi
 grep ^key_size entry
+grep ^key.\*count= entry | LANG=C sort
 
 echo
 echo '[Generating CSR.]'
@@ -151,6 +235,7 @@ if test "`grep ^state entry`" != state=HAVE_CSR ; then
 	grep ^state entry
 	exit 1
 fi
+grep ^key.\*count= entry | LANG=C sort
 
 echo
 echo '[Getting CSR signed.]'
@@ -160,15 +245,17 @@ if test "`grep ^state entry`" != state=NEED_TO_SAVE_CERT ; then
 	grep ^state entry
 	exit 1
 fi
+grep ^key.\*count= entry | LANG=C sort
 
 echo
 echo '[Saving certificate.]'
-$toolsdir/iterate ca entry START_SAVING_CERT,SAVING_CERT,NEED_TO_READ_CERT,READING_CERT,NEED_TO_SAVE_CA_CERTS,START_SAVING_CA_CERTS,SAVING_CA_CERTS,NEED_TO_NOTIFY_ISSUED_SAVED,NOTIFYING_ISSUED_SAVED,SAVED_CERT | sed 's@'"$tmpdir"'@$tmpdir@g'
+$toolsdir/iterate ca entry START_SAVING_CERT,PRE_SAVE_CERT,SAVING_CERT,NEED_TO_READ_CERT,READING_CERT,POST_SAVED_CERT,NEED_TO_SAVE_CA_CERTS,START_SAVING_CA_CERTS,SAVING_CA_CERTS,NEED_TO_NOTIFY_ISSUED_SAVED,NOTIFYING_ISSUED_SAVED,SAVED_CERT | sed 's@'"$tmpdir"'@$tmpdir@g'
 if test "`grep ^state entry`" != state=MONITORING ; then
 	echo Saving failed or did not move to monitoring.
 	grep ^state entry
 	exit 1
 fi
+grep ^key.\*count= entry | LANG=C sort
 
 echo
 echo '[From-scratch enrollment scenario OK.]'
@@ -293,9 +380,13 @@ ca_external_helper=$tmpdir/ca-issued
 EOF
 : > $tmpdir/certfile4
 $toolsdir/iterate ca3 entry3 NEED_KEYINFO,READING_KEYINFO,HAVE_KEYINFO
+grep ^key.\*count= entry3 | LANG=C sort
 $toolsdir/iterate ca3 entry3 NEED_CSR,GENERATING_CSR
+grep ^key.\*count= entry3 | LANG=C sort
 $toolsdir/iterate ca3 entry3 NEED_TO_SUBMIT,SUBMITTING
+grep ^key.\*count= entry3 | LANG=C sort
 $toolsdir/iterate ca3 entry3 NEED_TO_SAVE_CERT,SAVING_CERT,START_SAVING_CERT
+grep ^key.\*count= entry3 | LANG=C sort
 
 echo
 echo '[Enroll, helper produces noise before.]'
@@ -389,6 +480,85 @@ $toolsdir/iterate ca3 entry3 NEED_CSR,GENERATING_CSR
 $toolsdir/iterate ca3 entry3 NEED_TO_SUBMIT,SUBMITTING
 $toolsdir/iterate ca3 entry3 NEED_TO_SAVE_CERT,SAVING_CERT,START_SAVING_CERT
 
+echo
+echo '[Enroll, helper produces binary certificate output.]'
+cat > entry3 << EOF
+id=Test
+ca_name=Friendly
+state=HAVE_KEY_PAIR
+key_storage_type=FILE
+key_storage_location=$tmpdir/keyfile
+cert_storage_type=FILE
+cert_storage_location=$tmpdir/certfile4
+notification_method=STDOUT
+EOF
+cat > ca3 << EOF
+id=Friendly
+ca_type=EXTERNAL
+ca_external_helper=$tmpdir/ca-issued-binary-x509
+EOF
+: > $tmpdir/certfile4
+$toolsdir/iterate ca3 entry3 NEED_KEYINFO,READING_KEYINFO,HAVE_KEYINFO
+$toolsdir/iterate ca3 entry3 NEED_CSR,GENERATING_CSR
+$toolsdir/iterate ca3 entry3 NEED_TO_SUBMIT,SUBMITTING
+$toolsdir/iterate ca3 entry3 NEED_TO_SAVE_CERT,SAVING_CERT,START_SAVING_CERT
+
+echo
+echo '[Enroll, helper produces JSON output.]'
+cat > entry11 << EOF
+id=Test
+ca_name=Webby
+state=HAVE_KEY_PAIR
+key_storage_type=FILE
+key_storage_location=$tmpdir/keyfile
+cert_storage_type=FILE
+cert_storage_location=$tmpdir/certfile11
+root_cert_files=$tmpdir/rootcertfile
+other_cert_files=$tmpdir/othercertfile
+notification_method=STDOUT
+EOF
+cat > ca11 << EOF
+id=Webby
+ca_type=EXTERNAL
+ca_external_helper=$tmpdir/ca-issued-json
+EOF
+: > $tmpdir/certfile11
+rm -f $tmpdir/rootcertfile
+rm -f $tmpdir/othercertfile
+$toolsdir/iterate ca11 entry11 NEED_KEYINFO,READING_KEYINFO,HAVE_KEYINFO
+$toolsdir/iterate ca11 entry11 NEED_CSR,GENERATING_CSR
+$toolsdir/iterate ca11 entry11 NEED_TO_SUBMIT,SUBMITTING
+$toolsdir/iterate ca11 entry11 NEED_TO_SAVE_CERT,START_SAVING_CERT,SAVING_CERT,SAVED_CERT,NEED_TO_SAVE_CA_CERTS,START_SAVING_CA_CERTS,SAVING_CA_CERTS
+grep 'BEGIN CERTIFICATE' $tmpdir/certfile11 | wc -l
+grep 'BEGIN CERTIFICATE' $tmpdir/othercertfile | wc -l
+grep 'BEGIN CERTIFICATE' $tmpdir/rootcertfile | wc -l
+
+echo
+echo '[Try to enroll, but we need to generate a new key]'
+cat > entry12 << EOF
+id=Test
+ca_name=Webby
+state=HAVE_KEY_PAIR
+key_storage_type=FILE
+key_storage_location=$tmpdir/keyfile
+cert_storage_type=FILE
+cert_storage_location=$tmpdir/certfile12
+notification_method=STDOUT
+EOF
+cat > ca12 << EOF
+id=Webby
+ca_type=EXTERNAL
+ca_external_helper=$tmpdir/ca-need-rekey
+EOF
+: > $tmpdir/certfile12
+rm -f $tmpdir/rootcertfile
+rm -f $tmpdir/othercertfile
+$toolsdir/iterate ca12 entry12 NEED_KEYINFO,READING_KEYINFO,HAVE_KEYINFO
+$toolsdir/iterate ca12 entry12 NEED_CSR,GENERATING_CSR
+$toolsdir/iterate ca12 entry12 NEED_TO_SUBMIT,SUBMITTING
+$toolsdir/iterate ca12 entry12 NEED_KEY_PAIR,GENERATING_KEY_PAIR,HAVE_KEY_PAIR
+$toolsdir/iterate ca12 entry12 NEED_KEYINFO,READING_KEYINFO,HAVE_KEYINFO
+
 echo
 echo '[Enroll until we notice we have no specified CA.]'
 cat > entry3 << EOF
@@ -446,8 +616,36 @@ ca_type=EXTERNAL
 ca_external_helper=$tmpdir/ca-reject
 EOF
 $toolsdir/iterate ca5 entry5 NEED_KEYINFO,READING_KEYINFO,HAVE_KEYINFO
+grep ^key.\*count= entry5 | LANG=C sort
 $toolsdir/iterate ca5 entry5 NEED_CSR,GENERATING_CSR
+grep ^key.\*count= entry5 | LANG=C sort
 $toolsdir/iterate ca5 entry5 NEED_TO_SUBMIT,SUBMITTING
+grep ^key.\*count= entry5 | LANG=C sort
+$toolsdir/iterate ca5 entry5 NEED_TO_NOTIFY_REJECTION,NOTIFYING_REJECTION | sed 's@'"$tmpdir"'@$tmpdir@g'
+grep ^key.\*count= entry5 | LANG=C sort
+$toolsdir/iterate ca5 entry5 "" | sed 's@'"$tmpdir"'@$tmpdir@g'
+grep ^key.\*count= entry5 | LANG=C sort
+
+echo
+echo '[Enroll until the CA rejects us after poll.]'
+cat > entry5 << EOF
+id=Test
+ca_name=Meanie
+state=HAVE_KEY_PAIR
+key_storage_type=FILE
+key_storage_location=$tmpdir/keyfile
+cert_storage_type=FILE
+cert_storage_location=$tmpdir/certfile3
+notification_method=STDOUT
+EOF
+cat > ca5 << EOF
+id=Meanie
+ca_type=EXTERNAL
+ca_external_helper=$tmpdir/ca-reject-second-time
+EOF
+$toolsdir/iterate ca5 entry5 NEED_KEYINFO,READING_KEYINFO,HAVE_KEYINFO
+$toolsdir/iterate ca5 entry5 NEED_CSR,GENERATING_CSR
+$toolsdir/iterate ca5 entry5 NEED_TO_SUBMIT,SUBMITTING,CA_WORKING
 $toolsdir/iterate ca5 entry5 NEED_TO_NOTIFY_REJECTION,NOTIFYING_REJECTION | sed 's@'"$tmpdir"'@$tmpdir@g'
 $toolsdir/iterate ca5 entry5 "" | sed 's@'"$tmpdir"'@$tmpdir@g'
 
@@ -552,6 +750,43 @@ $toolsdir/iterate ca9 entry9 NEED_KEYINFO,READING_KEYINFO,HAVE_KEYINFO
 $toolsdir/iterate ca9 entry9 NEED_CSR,GENERATING_CSR
 $toolsdir/iterate ca9 entry9 NEED_TO_SUBMIT,SUBMITTING
 
+echo
+echo "[Enroll until we have SCEP data to go with it.]"
+cat > entry9 << EOF
+id=Test
+ca_name=SCEP
+state=HAVE_KEY_PAIR
+key_storage_type=FILE
+key_storage_location=$tmpdir/keyfile
+notification_method=STDOUT
+EOF
+cat > ca9 << EOF
+id=SCEP
+ca_type=EXTERNAL
+ca_external_helper=$tmpdir/ca-needs-scep-16
+ca_encryption_cert=-----BEGIN CERTIFICATE-----
+ MIICBDCCAW2gAwIBAgIEEjRWgTANBgkqhkiG9w0BAQUFADAaMRgwFgYDVQQDDA9U
+ ZXN0IExldmVsIDggQ0EwHhcNMTUwMjA0MTk0NjU4WhcNMTYwMjA0MTk0NjU4WjAX
+ MRUwEwYDVQQDDAxUZXN0IEVFIENlcnQwgZ8wDQYJKoZIhvcNAQEBBQADgY0AMIGJ
+ AoGBALjcinKYW+KHmciWmdXK5ZNRpKXcc6DqKykg0dUYgUsKTr6GYBeyA64Jmq8S
+ IOYqP2gWnSnw+LWQpbzKvCW0gCO6/skqwNdDZfcxXQmWVEE2oJPmu0a5I02DD46y
+ vVeugjriz2RHVNNjORXmf2xm6bZtcWtzzXew+H5lJIpRzj4LAgMBAAGjWjBYMAkG
+ A1UdEwQCMAAwCwYDVR0PBAQDAgTwMB0GA1UdDgQWBBRd3x1DMcHyzexXrenW0TRw
+ 3ANRyjAfBgNVHSMEGDAWgBQz4V1OzMt4ObAn9koy3aLP2bzFTjANBgkqhkiG9w0B
+ AQUFAAOBgQBozEcRs625HJ6YMZ2TLJKST1Z38ouIfwtl2Gv4WzGgVcRKVpoMgWjl
+ DbC+yjEDPm5+GwzEwVuR0E4g/nThfff/Ld8wVLfqdvClIUcgM8XEpPSRGrWLri+t
+ 9KqCx+t7heiWQcRD4OT1EfsHmXUz2+tAat6XvRcJ3AI1gtks0vJ6mA==
+ -----END CERTIFICATE-----
+EOF
+$toolsdir/iterate ca9 entry9 NEED_KEYINFO,READING_KEYINFO,HAVE_KEYINFO
+grep ^key.\*count= entry9 | LANG=C sort
+$toolsdir/iterate ca9 entry9 NEED_CSR,GENERATING_CSR
+grep ^key.\*count= entry9 | LANG=C sort
+$toolsdir/iterate ca9 entry9 NEED_TO_SUBMIT,SUBMITTING
+grep ^key.\*count= entry9 | LANG=C sort
+$toolsdir/iterate ca9 entry9 NEED_SCEP_DATA,GENERATING_SCEP_DATA,HAVE_SCEP_DATA
+grep ^key.\*count= entry9 | LANG=C sort
+
 # Note! The "iterate" harness rounds delay times up to the next multiple of 50.
 for interval in 0 30 1800 3600 7200 86000 86500 604800 1000000 2000000; do
 	now=`date +%s`
